/*
* This file is part of Firestorm NIDS
* Copyright (c) 2001 Gianni Tedesco
*
*    Firestorm is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    Firestorm is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU General Public License
*    along with Firestorm; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <s_globals.h>
#include <s_util.h>
#include <sensor.h>
#include <cleanup.h>
#include <targets.h>
#include <loader.h>
#include <capdev.h>
#include <demux.h>
#include <mem.h>

/* Our one and only global structure */
struct sensor_globals_t globals;

void sig_proc(int code)
{
	switch (code)
	{
		case SIGHUP:
		case SIGTERM:
		case SIGINT:
			fprintf(stderr, CRLF "Got signal %i, "
				"scheduling cleanup" CRLF , code);
			globals.keepalive.counter=0;
			return;
	}
}

/* Daemonize ourself and put ourselves in to
* a known state. Under UNIX processes inherit
* all kinds of crap, we need to obliterate it all */
int daemon_boogie(const char *lfn) 
{
	int lfd;
	int fd=0,fds;

	/* Close all file descriptors, we could
	* have inherited fds from parent processes etc.. */
	fds=sysconf(_SC_OPEN_MAX);
	while (fd<fds) close(fd++);
	
	/* Set our umask, this is inherited */
	umask(0);
	
	/* First fork off in to 
	* the background */
	switch( fork() ) {
		case 0: break;
		case -1: return 0;
		default: _exit(0);
	}

	/* Create a new session and become
	* session group leader */
	if ( setsid() < 0 ) {
		return 0;
	}

	/* Fork once again, we now loose
	* control of the session and can
	* never regain a tty ever again */
	switch( fork() ) {
		case 0: break;
		case -1: return 0;
		default: _exit(0);
	}

	/* Open our log file */
	lfd=open(lfn, O_WRONLY | O_CREAT, 00600);
	if ( lfd < 0 ) return 0;

	/* dup stdout to the log file */
	if ( dup2(lfd, 1) < 0 ) return 0;
	
	return 1;
}

const char *plugin_get_var(const char *name)
{
	p_kv t;

	if ( !name ) return NULL;
	
	for(t=globals.plugin_vars; t; t=(p_kv)t->l.next) {
		if ( !strcasecmp(name, t->key) ) {
			return t->val;
		}
	}
	
	return NULL;
}

void __inline__ clean_packet(p_packet p)
{
	memset(p, 0, sizeof(packet_t));
}

void __inline__ packet_top_layers(p_packet p) {
	int i;

	for ( i=0; i < p->top_layer; i++ ) {
		switch ( p->layer[i].type ) {
			case LTYPE_LINK:
				p->top_link=i;
				break;
			case LTYPE_NET:
				p->top_net=i;
				break;
			case LTYPE_TRANS:
				p->top_trans=i;
				break;
			case LTYPE_APP:
				p->top_app=i;
				break;
			case LTYPE_ENCAP:
				p->top_encap=i;
				break;
		}
	}
}

void *capture_proc(void *parm)
{
	p_capdev c;
	p_packet p;
	struct tm *tmpt;
	uint a=0;
#ifdef __DEBUG_DECODE
	int i;
#endif
	
	c=(p_capdev)parm;
	c->state=CAP_STATE_CAPTURE;
	
	while(1){

		/* Poll the capture device */
		if ( (p=c->capture->poll(c->priv)) == NULL ) break;
		
		/* Convert the time to local time*/
		CRITICAL_START(globals.tlock);
		tmpt = localtime(&p->time);
		memcpy(&p->ltime, tmpt, sizeof(p->ltime));
		CRITICAL_STOP(globals.tlock);
		
		/* Tell the packet who cought it */
		p->capdev=c;

		/* Demiltiplex... */
		p->top_layer=demux(p);
		packet_top_layers(p);
		
#ifdef __DEBUG_DECODE
		print_out("[**] DEBUG PACKET [**] - %i layers"
			CRLF, p->top_layer+1);
		for (i=0; i<p->top_layer+1;i++){
			print_out("LAYER=%i TYPE=%i ID=%i(0x%X) :: %s" CRLF,
				i, p->layer[i].type,
				p->layer[i].id, p->layer[i].id,
				p->layer[i].name);
		}print_raw(CRLF);a++; continue;
#endif

		/* Match */
		matcher_begin(p);

		/* Cleanup the packet so stale data 
		* isn't left over for next time we poll */
		clean_packet(p);

		/* Per capdev packet count */
		a++;
	}

	print_out("Ending %s capture [%s]: Caught %u packets." CRLF, 
		c->capture->name, c->interface, a);
	
	c->capture->end(c->priv);
	c->state=CAP_STATE_NULL;
	
	atomic_dec_counter(&globals.keepalive);
	return;	
}

int main ( int argc, char **argv ) 
{
	/* Curse me for this needless waste of stack space in main() */
	char *lfn;

	memset(&globals, 0, sizeof(globals));
	tzset();
	
#ifdef LEAK_CHECK
	if ( !mem_init () ) return EXIT_ERR;
	globals.malloc=fs_alloc;
	globals.free=fs_free;
	globals.fs_strdup=fs_strdup;
#else
	globals.malloc=malloc;
	globals.free=free;
	globals.fs_strdup=strdup;
#endif

	print_raw(VERSION_STRING);

	if ( argc < 2 || argc > 3) {
		print_out("Usage: %s <configfile> [<logfile>]" CRLF, argv[0]);
		return EXIT_ERR;
	}else{
		globals.config_fn=argv[1];
	}
	
	/* Can't go nowhere with no cleanup subsystem */
	if ( !cleanup_init() ) return EXIT_ERR;

#if 1
	/* Do the daemon boogie */
	lfn=(argc == 3) ? argv[2] : "firestorm.log";
	if ( !daemon_boogie(lfn) ) {
		cleanup(EXIT_ERR);
	}
#endif

	/* Also print version info to file */
	print_raw(VERSION_STRING CRLF);
	
	/* Create global mutex, and add cleanup code */
	MUTEX_CREATE(globals.glock, "Global");
	MUTEX_CREATE(globals.tlock, "localtime()");
	MUTEX_CREATE(globals.keepalive.lock, "Keepalive");
	globals.keepalive.counter=0;
	cleanup_add_handler(cleanup_lock, (void *)&globals.glock);
	cleanup_add_handler(cleanup_lock, (void *)&globals.tlock);
	cleanup_add_handler(cleanup_lock, (void *)&globals.keepalive.lock);
	
	/* Set signal handlers */
	signal(SIGTERM, sig_proc);
	signal(SIGINT, sig_proc);
	signal(SIGHUP, sig_proc);

	
	/* Load capture subsystem */
	if ( !capdev_init() ) cleanup(EXIT_ERR);

	/* Load demultiplexing subsystem */
	if ( !demux_init() ) cleanup(EXIT_ERR);
	
	/* Load packet target subsystem */
	if ( !target_init() ) cleanup(EXIT_ERR);
	
	/* Load packet matcher subsystem  */
	if ( !matcher_init() ) cleanup(EXIT_ERR);
	
	/* Load ruleset file subsytem  */
	if ( !rules_init() ) cleanup(EXIT_ERR);

	/* Load and run our config */
	if ( !conf_load_config(globals.config_fn) ) {
		conf_cleanup_temp();
		cleanup(EXIT_ERR);
	}
	if ( !conf_go() ) {
		conf_cleanup_temp();
		cleanup(EXIT_ERR);
	}conf_cleanup_temp();

	/* Print a quick summary */
	print_out("PLUGINS: %u plugins loaded, "
		"%u matchers, "
		"%u targets." CRLF,
		globals.num_plugins,
		globals.num_matchers,
		globals.num_targets);
	
	print_out("RULESET: %u rule files loaded, "
		"%u rules all together." CRLF,
		globals.num_rulefiles,
		globals.num_rules);
	
#ifdef LEAK_CHECK
	/* Display dynamic memory statistics */
	fs_usage();
#endif
	
	/* Now we start capturing for real */
	capdev_start_all();
	
	while(globals.keepalive.counter) {
		/* I guess we would do regular maintenance
		* crap here at some point  */
		sleep(1);
	}
	
	cleanup(EXIT_OK);
}
